// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.
#ifndef NUKE_MOV64READER_H
#define NUKE_MOV64READER_H

#include "Build/fnBuild.h"
#include "DDImage/DDString.h"
#include "DDImage/Reader.h"
#include "DDImage/Row.h"
#include "DDImage/MemoryBuffer.h"
#include "DDImage/RefCountedObject.h"
#include "DDImage/Thread.h"
#include "DDImage/LUT.h"

#include "codecWhitelist.h"

#include "mov64Common.h"
#define NUKE_MOV64READER_CPP
#include "movCommon.h"

#define NUKE_MOV64_PLANAR

#ifdef _WIN32
  #include <io.h>
#endif

#ifndef INT64_C
# define INT64_C(c) (c ## LL)
#endif
#ifndef UINT64_C
# define UINT64_C(c) (c ## ULL)
#endif

extern "C" {
#include <errno.h>
#include <libavformat/avformat.h>
#include <libavutil/imgutils.h>
#include <libavformat/avio.h>
#include <libavcodec/avcodec.h>
#include <libavutil/opt.h>
#include <libswscale/swscale.h>
#include <libavutil/avutil.h>
#include <libavutil/error.h>
}

////////////////////////////////////////////////////////////////////////////////
// FN_LICENSED_PRORES_DNXHD_CODEC
// Comment out this preprocessor definition if building this file WITHOUT a
// valid Apple ProRes license or valid Avid DNxHD license.
#define FN_LICENSED_PRORES_DNXHD_CODEC

////////////////////////////////////////////////////////////////////////////////

#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
# include "codecProResCommon.h" // String definitions of the codec profile names.
# include <NukeCodecs/fnNukeCodecs.h>
#endif

// Set non-zero to enable tracing of file information inspected while opening a file in FFmpegFile::FFmpegFile().
// Make sure this is disabled before checking-in this file.
#define TRACE_FILE_OPEN 0

// Set non-zero to enable tracing of FFmpegFile::decode() general processing (plus error reports).
// Make sure this is disabled before checking-in this file.
#define TRACE_DECODE_PROCESS 0

// Set non-zero to enable tracing of the first few bytes of each data block in the bitstream for each frame decoded. This
// assumes a 4-byte big-endian byte count at the start of each data block, followed by that many bytes of data. There may be
// multiple data blocks per frame. Each data block represents an atomic unit of data input to the decoder (e.g. an H.264 NALU).
// It works nicely for H.264 streams in .mov files but there's no guarantee for any other format. This can be useful if you
// need to see what kind of NALUs are being decoded.
// Make sure this is disabled before checking-in this file.
#define TRACE_DECODE_BITSTREAM 0

// Subset of Foundry::Colour::YCbCrEncodingEnum
enum YCbCrEncodingEnum {
  eYCbCrEncodingRec709,
  eYCbCrEncodingRec601,
  eNumYCbCrEncodings
};

const char* const YCbCrEncodingNames[eNumYCbCrEncodings + 1] =
{
  "Rec 709",
  "Rec 601",
  0 // NULL terminator
};


class FFmpegFile : public DD::Image::RefCountedObject
{
  struct Stream
  {
    int _idx;                      // stream index
    AVStream* _avstream;           // video stream
    AVCodecContext* _codecContext; // video codec context
    AVCodec* _videoCodec;
    AVFrame* _avFrame;             // decoding frame
    SwsContext* _convertCtx;
    bool _resetConvertCtx;

    int _fpsNum;
    int _fpsDen;

    int64_t _startPTS;     // PTS of the first frame in the stream
    int64_t _frames;       // video duration in frames

    bool _ptsSeen;                      // True if a read AVPacket has ever contained a valid PTS during this stream's decode,
    // indicating that this stream does contain PTSs.
    int64_t AVPacket::*_timestampField; // Pointer to member of AVPacket from which timestamps are to be retrieved. Enables
    // fallback to using DTSs for a stream if PTSs turn out not to be available.

    int                 _width;
    int                 _height;
    double              _aspect;
    int                 _bitDepth;
    int                 _numberOfComponents;
    enum AVPixelFormat  _outputPixelFormat;
    uint8_t             _componentPosition[4];
    int                 _colorMatrixTypeOverride;       // Option to override the default YCbCr color matrix. 0 means no override (default).
    bool                _doNotAttachPrefix;
    bool                _matchMetaFormat;

    int _decodeNextFrameIn; // The 0-based index of the next frame to be fed into decode. Negative before any
    // frames have been decoded or when we've just seeked but not yet found a relevant frame. Equal to
    // frames_ when all available frames have been fed into decode.

    int _decodeNextFrameOut; // The 0-based index of the next frame expected out of decode. Negative before
    // any frames have been decoded or when we've just seeked but not yet found a relevant frame. Equal to
    // frames_ when all available frames have been output from decode.

    int _accumDecodeLatency; // The number of frames that have been input without any frame being output so far in this stream
    // since the last seek. This is part of a guard mechanism to detect when decode appears to have
    // stalled and ensure that FFmpegFile::decode() does not loop indefinitely.

    Stream()
      : _idx(0)
      , _avstream(NULL)
      , _codecContext(NULL)
      , _videoCodec(NULL)
      , _avFrame(NULL)
      , _convertCtx(NULL)
      , _resetConvertCtx(true)
      , _fpsNum(1)
      , _fpsDen(1)
      , _startPTS(0)
      , _frames(0)
      , _ptsSeen(false)
      , _timestampField(&AVPacket::pts)
      , _width(0)
      , _height(0)
      , _aspect(1.0)
      , _bitDepth(8)
      , _numberOfComponents(3)
      , _outputPixelFormat(AV_PIX_FMT_RGB24)
      , _colorMatrixTypeOverride(0)
      , _decodeNextFrameIn(-1)
      , _decodeNextFrameOut(-1)
      , _accumDecodeLatency(0)
      , _doNotAttachPrefix(true)
      , _matchMetaFormat(true)
    {
      // The purpose of this is to avoid an RGB->RGB conversion.
      // This saves memory and improves performance. For example
      // Nuke expects RGBA component order. So for any RGB pixel
      // format which does not have this ordering, the following
      // should be used to remap the components, e.g. ARGB->RGBA,
      // BGRA->RGBA, BGR->RGB, etc. The output of the DNxHD decoder
      // is an example of this in action.
      _componentPosition[0] = 0; // Red position.
      _componentPosition[1] = 1; // Green position.
      _componentPosition[2] = 2; // Blue position.
      _componentPosition[3] = 3; // Alpha position.
    }

    ~Stream()
    {

      if (_avFrame)
        av_free(_avFrame);

      if (_codecContext)
        avcodec_close(_codecContext);

      if (_convertCtx)
        sws_freeContext(_convertCtx);
    }

    static void destroy(Stream* s)
    {
      delete(s);
    }

    int64_t frameToPts(int frame) const
    {
      int64_t numerator = int64_t(frame) * _fpsDen *  _avstream->time_base.den;
      int64_t denominator = int64_t(_fpsNum) * _avstream->time_base.num;
      return _startPTS + (numerator / denominator);
    }

    int ptsToFrame(int64_t pts) const
    {
      int64_t numerator = int64_t(pts - _startPTS) * _avstream->time_base.num * _fpsNum;
      int64_t denominator = int64_t(_avstream->time_base.den) * _fpsDen;
      return static_cast<int>(numerator / denominator);
    }

    bool isRec709Format()
    {
      // First check for codecs which require special handling:
      //  * JPEG codecs always use Rec 601.
      AVCodecID codecID = _codecContext->codec_id;

      if (codecID == AV_CODEC_ID_MJPEG )  {
        return false;
      }

      // Using method described in step 5 of QuickTimeCodecReader::setPreferredMetadata
      return (_height >= 720);
    }

    bool isYUV() const
    {
      // from swscale_internal.h
      const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(_codecContext->pix_fmt);
      return !(desc->flags & AV_PIX_FMT_FLAG_RGB) && desc->nb_components >= 2;
    }

    static float GetStreamAspectRatio(Stream* stream);
    
    // Generate the conversion context used by SoftWareScaler if not already set.
    // |reset| forces recalculation of cached context.
    SwsContext* getConvertCtx(AVPixelFormat srcPixelFormat, int srcWidth, int srcHeight, int srcColorRange, AVPixelFormat dstPixelFormat, int dstWidth, int dstHeight);

    // Return the number of input frames needed by this stream's codec before it can produce output. We expect to have to
    // wait this many frames to receive output; any more and a decode stall is detected.
    // In FFmpeg 2.1.4, I found that some codecs now support multithreaded decode which appears as latency
    // I needed to add the thread_count onto the codec delay - pickles
    int getCodecDelay() const
    {
      return ((_videoCodec->capabilities & CODEC_CAP_DELAY) ? _codecContext->delay : 0)
        + _codecContext->has_b_frames
        + _codecContext->thread_count;
    }
  };


  // AV structure
  AVFormatContext* _context;
  AVInputFormat*   _format;

  // store all video streams available in the file
  std::vector<Stream*> _streams;

  // reader error state
  std::string _errorMsg;  // internal decoding error string
  bool _invalidState;     // true if the reader is in an invalid state

  AVPacket _avPacket;

  // internal lock for multithread access
  DD::Image::Lock _lock;

#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
  // Apple ProRes and Avid DNxHD codec specific.
  std::auto_ptr<NukeCodecs::IDecoder> _decoder;
  NukeCodecs::CodecInfo _codecInfo;
#endif

  // set reader error
  void setError(const char* msg, const char* prefix = 0)
  {
    if (prefix) {
      _errorMsg = prefix;
      _errorMsg += msg;
#if TRACE_DECODE_PROCESS
      std::cout << "!!ERROR: " << prefix << msg << std::endl;
#endif
    }
    else {
      _errorMsg = msg;
#if TRACE_DECODE_PROCESS
      std::cout << "!!ERROR: " << msg << std::endl;
#endif
    }
    _invalidState = true;
  }

  // set FFmpeg library error
  void setInternalError(const int error, const char* prefix = 0)
  {
    char errorBuf[1024];
    av_strerror(error, errorBuf, sizeof(errorBuf));
    setError(errorBuf, prefix);
  }

  // get stream start time
  int64_t getStreamStartTime(Stream& stream);

  // Get the video stream duration in frames...
  int64_t getStreamFrames(Stream& stream);

  ///////////////////////////////////////////////////////////////////////////////////////
public:

  typedef DD::Image::RefCountedPtr<FFmpegFile> Ptr;

  // constructor
  FFmpegFile(const char* filename);

  // destructor
  ~FFmpegFile();

  // get the internal error string
  const char* error() const
  {
    return _errorMsg.c_str();
  }

  // return true if the reader can't decode the frame
  bool invalid() const
  {
    return _invalidState;
  }

  // return the numbers of streams supported by the reader
  size_t streams() const
  {
    return _streams.size();
  }

  void setColorMatrixTypeOverride(int colorMatrixType) const
  {
    if (_streams.empty()) {
      return;
    }

    // mov64Reader::decode always uses stream 0
    Stream* stream = _streams[0];
    stream->_colorMatrixTypeOverride = colorMatrixType;
    stream->_resetConvertCtx = true;
  }
  
  void setDoNotAttachPrefix(bool doNotAttachPrefix) const
  {
    if (_streams.empty()) {
      return;
    }
    
    // mov64Reader::decode always uses stream 0
    Stream* stream = _streams[0];
    stream->_doNotAttachPrefix = doNotAttachPrefix;
  }

  void setMatchMetaFormat(bool matchMetaFormat) const
  {
    if (_streams.empty()) {
      return;
    }
    
    // mov64Reader::decode always uses stream 0
    Stream* stream = _streams[0];
    stream->_matchMetaFormat = matchMetaFormat;
  }
  
  bool isRec709Format() const
  {
    if (_streams.empty()) {
      return false;
    }

    // mov64Reader::decode always uses stream 0
    Stream* stream = _streams[0];
    return stream->isRec709Format();
  }

  bool isYUV() const
  {
    if (_streams.empty()) {
      return false;
    }

    // mov64Reader::decode always uses stream 0
    Stream* stream = _streams[0];
    return stream->isYUV();
  }

  int getBitDepth() const {
    if (_streams.empty()) {
      return 0;
    }

    // mov64Reader::decode always uses stream 0
    // Sometimes AVCodec reports a bitdepth of 0 (eg PNG codec).
    // In this case, assume 8 bit.
    return std::max(8, _streams[0]->_bitDepth);
  }

  int getNumberOfComponents() const {
    if (_streams.empty()) {
      return 0;
    }
    // mov64Reader::decode always uses stream 0
    return _streams[0]->_numberOfComponents;
  }

  int getComponentPosition(DD::Image::Channel channel) const {
    if (_streams.empty()) {
      return 0;
    }

    int componentIndex;
    if (DD::Image::Chan_Red == channel)
      componentIndex = 0;
    else if (DD::Image::Chan_Green == channel)
      componentIndex = 1;
    else if (DD::Image::Chan_Blue == channel)
      componentIndex = 2;
    else // Assume alpha.
      componentIndex = 3;
    // mov64Reader::decode always uses stream 0
    return _streams[0]->_componentPosition[componentIndex];
  }

  int getComponentPosition(DD::Image::ChannelSetInit mask) const {
    if (_streams.empty()) {
      return 0;
    }

    int componentIndex;
    if (DD::Image::Mask_Red == mask)
      componentIndex = 0;
    else if (DD::Image::Mask_Green == mask)
      componentIndex = 1;
    else if (DD::Image::Mask_Blue == mask)
      componentIndex = 2;
    else // Assume alpha.
      componentIndex = 3;
    // mov64Reader::decode always uses stream 0
    return _streams[0]->_componentPosition[componentIndex];
  }

  int getWidth() const {
    if (_streams.empty()) {
      return 0;
    }

    // mov64Reader::decode always uses stream 0
    return _streams[0]->_width;
  }

  int getHeight() const {
    if (_streams.empty()) {
      return 0;
    }

    // mov64Reader::decode always uses stream 0
    return _streams[0]->_height;
  }

  // decode a single frame into the buffer. Thread safe
  bool decode(unsigned char* buffer, unsigned frame, size_t streamIdx = 0);

  // fill the metada information of a particular stream
  bool metadata(DD::Image::MetaData::Bundle& metadata, size_t streamIdx = 0);

  // get stream information
  bool info( int& width,
             int& height,
             double& aspect,
             int& frames,
             size_t streamIdx = 0);

  const char* getColorspace() const;

  int getBufferSize() const;
  int getRowSize() const;
};

//////////////////////////////////////////////////////////////////////////////
// Keeps track of all FFmpegFile mapped against file name.
class FFmpegFileManager
{
  typedef std::map<std::string, FFmpegFile::Ptr> ReaderMap;

#if !defined(FN_HIERO)
  ReaderMap _readers;

  // internal lock
  DD::Image::Lock _lock;
#endif

  // A lock manager function for FFmpeg, enabling it to use mutexes managed by this reader. Pass to av_lockmgr_register().
  static int FFmpegLockManager(void** mutex, enum AVLockOp op);

public:

  // singleton
  static FFmpegFileManager s_readerManager;

  // constructor
  FFmpegFileManager();

  // get a specific reader
  FFmpegFile::Ptr get(const char* filename);

  // release a specific reader
  void release(const char* filename);
};

/////////////////////////////////////////////////////////////////////////////

class mov64Reader : public DD::Image::Reader, public DD::Image::IBufferFillHeap
{
public:
  // If supplied, optionalFilename will be used over the iops FileKnob.
  // This allows the movReader to perform an early check. See invalid().
  explicit mov64Reader(DD::Image::Read* iop, const char* optionalFilename = NULL);
  ~mov64Reader();

  // Expose protected function for use by movReader
  virtual void prefetchMetaData() { DD::Image::Reader::prefetchMetaData(); }
  const DD::Image::MetaData::Bundle& fetchMetaData(const char* key) { return meta; }
  virtual bool videosequence() const { return true; }
  void engine(int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row);
  void open();

#ifdef NUKE_MOV64_PLANAR
  // Planar reader extensions
  virtual DD::Image::PlanarReadInfo planarReadInfo(const DD::Image::ChannelSet& channels);
  virtual void planarReadAndDecode(DD::Image::GenericImagePlane& image,
                                   const DD::Image::ChannelSet& channels);
  virtual int planarReadPass(void* buffer,
                             const DD::Image::ChannelSet& channels);
  virtual void planarDecodePass(void* srcBuffer,
                                DD::Image::GenericImagePlane& image,
                                const DD::Image::ChannelSet& channels,
                                int threadIndex,
                                int nDecodeThreads);
#endif // NUKE_MOV64_PLANAR

  // For movReader, to query the ability of mov64 to read a particular codec
  bool invalid();

private:

  // redefined from DD::Image::IBufferFill
  virtual bool fillBuffer(void* data);

  virtual void memoryInfo(DD::Image::Memory::MemoryInfoArray& output, const void* restrict_to) const;

  void engine8( int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row, unsigned char* data);
  void engine16(int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row, U16* data);
private:

  FFmpegFile::Ptr _reader;
  DD::Image::MemoryBuffer _data;    // decoding buffer
  size_t _memNeeded;                // memory needed for decoding a single frame

  int  _numFrames;                  // number of frames in the selected stream
  int  _currentFrame;               // current decoding frame

  DD::Image::MetaData::Bundle meta;
  static const Reader::Description d;

  const char* const _optionalFilename;

  uint8_t* _planarReadBuffer;
};

/////////////////////////////////////////////////////////
class mov64ReaderFormat : public DD::Image::ReaderFormat
{
  friend class mov64Reader;

public:
  mov64ReaderFormat(DD::Image::Read* iop);

  void knobs(DD::Image::Knob_Callback c);
  void append(DD::Image::Hash& hash);
  void initDynamicKnobs();

  int  getColorMatrixType()   const { return _colorMatrixType; }
  bool getDoNotAttachPrefix() const { return _doNotAttachPrefix; }
  bool getMatchMetaFormat()   const { return _matchMetaFormat; }
  void setDefaultColorMatrixType(int colorMatrixType)  { _defaultColorMatrixType = colorMatrixType; }
  void setYCbCrMatrixEnabled(bool enabled);

private:
  void updateColorMatrixKnob(DD::Image::Knob* colorMatrixKnob);

private:
  DD::Image::Read* _iop;
  int _colorMatrixType;
  int _defaultColorMatrixType;
  bool _doNotAttachPrefix;
  bool _matchMetaFormat;

  // Based on Mov32Reader
  static const int kNumDefaultLabels = 1;
  static const int kNumListDelimiters = 1;
  static const int kNumColorMatrixLabels = kNumDefaultLabels + eNumYCbCrEncodings + kNumListDelimiters;
  const char* _colourMatrixTypeLabels[kNumColorMatrixLabels];
};

#endif // NUKE_MOV64READER_H

