// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.

#ifndef NUKE_MOV64WRITER_H
#define NUKE_MOV64WRITER_H

#include "DDImage/Writer.h"
#include <boost/shared_ptr.hpp>

#ifndef INT64_C
# define INT64_C(c) (c ## LL)
#endif
#ifndef UINT64_C
# define UINT64_C(c) (c ## ULL)
#endif

#include "mov64Common.h"
#define NUKE_MOV64WRITER_CPP
#include "movCommon.h"

extern "C" {
#include <errno.h>
#include <libavformat/avformat.h>
#include <libavcodec/avcodec.h>
#include <libavutil/imgutils.h>
#include <libavformat/avio.h>
#include <libavutil/opt.h>
#include <libswscale/swscale.h>
#include <libavutil/avutil.h>
#include <libavutil/error.h>
#include <libswresample/swresample.h>
#include <libavutil/intreadwrite.h>
}

////////////////////////////////////////////////////////////////////////////////
// FN_LICENSED_PRORES_CODEC
// Comment out this preprocessor definition if building this file WITHOUT a
// valid Apple ProRes license.
#define FN_LICENSED_PRORES_CODEC

////////////////////////////////////////////////////////////////////////////////

#if defined(FN_LICENSED_PRORES_CODEC)
# include "codecProResCommon.h" // String definitions of the codec profile names.
# include <NukeCodecs/fnNukeCodecs.h>
# define SIZEOF_ARRAY(array) (sizeof(array) / sizeof(array[0]))

////////////////////////////////////////////////////////////////////////////////
// Apple ProRes codec support.
// This is a brief guide to the changes for adding Apple ProRes encoding to the
// mov64 writer.
// * Uses existing mov64 writer configuration code.
// * Overall, configure the writer as if the non SDK ProRes implementation were
//   to be used for encoding.
// * Tweak the Nuke pixel format -> encoder pixel format so that the encoder
//   pixel format is one that can be used with the Apple ProRes library.
// * At the point of encoding, the Apple ProRes library is used in place of the
//   FFmpeg avcodec_encode_video2 call.
//
// Encoding data involves converting from floating point RGB (Nuke) to the input
// pixel format of the encoding, then encoding and finally writing the encoded
// bit stream to the file.
//
// One of the main issues is that non SDK almost exclusively supports planar
// formats (lots at 8-bit and support rapidly dropping off for 10-bit and
// higher). It has very limited support for packed formats which are the ONLY
// formats that the Apple ProRes encoder accepts. So to use the Apple ProRes
// encoder library, manual pixel format conversions (packing) are provided.
//
// For a licensed Apple ProRes codec, define FN_LICENSED_PRORES_CODEC above.
//
////////////////////////////////////////////////////////////////////////////////
// AppleProResCodecStr
// What's all this about?
// The official ProRes codec library from Apple has been licensed and is used
// here to provide the encoding.
// In order to get this to work the Apple ProRes encoder library is used when
// the uncompressed frame is encoded. Up to that point, the non SDK libraries
// are initialised and configured to use the non SDK ProRes encoder. This has
// been done to keep things simple in order to minimise regression issues, use
// all the existing code, e.g. knobs, colourspace selection etc.
// This structure is used to define Apple ProRes specific parameters, such as
// the input pixel format that is required by the Apple ProRes codec libary.
struct AppleProResCodecStr
{
  const char* longName; // User readable description of the codec.
  const char* shortName; // Short name for faster internal parsing
  NukeCodecs::CodecFactory::CodecType codecType; // The ProRes encoder profile (Apple FCC). This is also the FCC used to identify the codec profile in the QuickTime movie.
  AVPixelFormat avPixelFormat; // The input pixel format to the encoder (FFmpeg definition).
  const char* knobLabel; // Knob label in the format "fourcc\tUserLabel"
};
typedef struct AppleProResCodecStr AppleProResCodecStr;

static const AppleProResCodecStr appleProResCodecLUT[] = {
//  { APPLE_PRORES_4444XQ_NAME, "fn_ap4x", NukeCodecs::CodecFactory::eProResCodec4444XQ, AV_PIX_FMT_BGRA64LE, "ap4x\tApple ProRes 4444 XQ" },
  { APPLE_PRORES_4444_NAME, "fn_ap4h", NukeCodecs::CodecFactory::eProResCodec4444, AV_PIX_FMT_BGRA64LE, "ap4h\tApple ProRes 4444" },
  { APPLE_PRORES_422HQ_NAME, "fn_apch", NukeCodecs::CodecFactory::eProResCodec422HQ, AV_PIX_FMT_BGRA64LE, "apch\tApple ProRes 422 HQ" },
  { APPLE_PRORES_422_NAME, "fn_apcn", NukeCodecs::CodecFactory::eProResCodec422, AV_PIX_FMT_BGRA64LE, "apcn\tApple ProRes 422" },
  { APPLE_PRORES_422LT_NAME, "fn_apcs", NukeCodecs::CodecFactory::eProResCodec422LT, AV_PIX_FMT_BGRA64LE, "apcs\tApple ProRes 422 LT" },
  { APPLE_PRORES_422PROXY_NAME, "fn_apco", NukeCodecs::CodecFactory::eProResCodec422Proxy, AV_PIX_FMT_BGRA64LE, "apco\tApple ProRes 422 Proxy" },
};

////////////////////////////////////////////////////////////////////////////////
#endif // FN_LICENSED_PRORES_CODEC

#define DEBUG_AUDIO_READER(x) std::cout << x << std::endl
//#define DEBUG_AUDIO_READER(x) ;


////////////////////////////////////////////////////////////////////////////////
// FnAVFrame
// Wrapper class for the FFmpeg AVFrame structure.
// In order negate the chances of memory leaks when using FFmpeg this class
// provides the initialisation, allocation and release of memory associated
// with an AVFrame structure.
//
// Note that this has been designed to be a drop in replacement for host
// managed memory for colourspace, pixel format and sample format conversions.
// It is not designed to be used with avcodec_decode_video2 as the underlying
// decoder usually manages memory.
//
// Example usage:
//
// Audio
//
//    FnAVFrame avFrame;
//    ret = avFrame.alloc(channels, nb_samples, _avSampleFormat, 1);
//    :
//    ret = swr_convert(_swrContext, avFrame->data, ...);
//    :
//    ret = avcodec_encode_audio2(avCodecContext, &pkt, avFrame, &gotPacket);
//    :
//
// Video
//
//    FnAVFrame avFrame;
//    ret = avFrame.alloc(width(), height(), pixelFormatCodec, 1);
//    if (!ret) {
//      :
//      sws_scale(convertCtx, ..., avFrame->data, avFrame->linesize);
//      :
//    }
//
// IMPORTANT
// This class has been purposefully designed NOT to have parameterised
// constructors or assignment operators. The reason for this is due to the
// complexity of managing the lifetime of the structure and its associated
// memory buffers.
//
class FnAVFrame
{
public:
  ////////////////////////////////////////////////////////////////////////////////
  // Ctor. Initialise the AVFrame structure.
  FnAVFrame()
    : _avFrame(NULL)
  {
  }

  ////////////////////////////////////////////////////////////////////////////////
  // Dtor. Release all resources. Release the AVFrame buffers and the
  // AVFrame structure.
  ~FnAVFrame();

  ////////////////////////////////////////////////////////////////////////////////
  // alloc
  // VIDEO SPECIFIC.
  // Allocate a buffer or buffers for the AVFrame structure. How many
  // buffers depends upon |avPixelFormat| which is a VIDEO format.
  //
  // @param width Frame width in pixels.
  // @param height Frame height in pixels.
  // @param avPixelFormat An AVPixelFormat enumeration for the frame pixel format.
  // @param align Buffer byte alignment.
  //
  // @return 0 if successful.
  //         <0 otherwise.
  //
  int alloc(int width, int height, enum AVPixelFormat avPixelFormat, int align);

  ////////////////////////////////////////////////////////////////////////////////
  // alloc
  // AUDIO SPECIFIC.
  // Allocate a buffer or buffers for the AVFrame structure. How many
  // buffers depends upon |avSampleFormat| which is an AUDIO format.
  //
  // @param nbChannels The number of audio channels.
  // @param nbSamples The number of audio samples that the buffer will hold.
  // @param avSampleFormat An AVSampleFormat enumeration for the audio format.
  // @param align Buffer byte alignment.
  //
  // @return 0 if successful.
  //         <0 otherwise.
  //
  int alloc(int nbChannels, int nbSamples, enum AVSampleFormat avSampleFormat, int align);

  // operator dereference overload.
  AVFrame* operator->() const { return _avFrame; }
  // operator type cast overload.
  operator AVFrame*() { return _avFrame; }

private:
  AVFrame* _avFrame;

  // Release any memory allocated to the data member variable of
  // the AVFrame structure.
  void deallocateAVFrameData();

  // Probably do not need the following or the headaches
  // of trying to support the functionality.
  // Hide the copy constuctor. Who would manage memory?
  FnAVFrame(FnAVFrame& avFrame) { }
  // Hide the assignment operator. Who would manage memory?
  FnAVFrame& operator=(const FnAVFrame& rhs) { return *this; }
};

////////////////////////////////////////////////////////////////////////////////
// AudioReader
// Class that uses the avformat, avcodec, avutil and swresample libraries to
// read, decode and reformat audio data from a file that contains audio.
//
// This class is used if an optional audio file has been specified to be used
// with the mov64Writer. The audio data is interleaved with the video frames
// by the mov64Writer.
//
// The mov64Writer can specify an 'offset' for the audio in order to advance
// or delay the audio with respect to the video. If the audio is delayed then
// the reader will create silence, keep track of the read position and provide
// 'real' audio at the appropriate point. If the audio is advanced then a
// seek operation is performed into the file and reads will begin from that
// position.
//
class AudioReader
{
public:
  AudioReader()
    : _avFormatContext(NULL)
    , _avStreamAudio(NULL)
    , _avFrameAudio(NULL)
    , _swrContext(NULL)
    , _avSampleFormat(AV_SAMPLE_FMT_S16)
    , _readPosition(0)
  {
  }

  ~AudioReader()
  {
    close();
  }

  ////////////////////////////////////////////////////////////////////////////////
  // open
  // Attempt to open the file and determine whether the file contains any audio
  // and whether that audio can be decoded.
  //
  // @param filename The name of the file containing audio.
  //
  // @return 0 if successful, non-zero otherwise.
  //
  int open(std::string filename);

  ////////////////////////////////////////////////////////////////////////////////
  // read
  // Read, decode and resample some audio data from the source file. The format
  // of the data will be in the desired target format.
  //
  // @param avFrame A reference to an FnAVFrame object. This is an input and
  //                output parameter. An UNINITIALIZED FnAVFrame is passed to
  //                this method which allocs an appropriately sized buffer. The
  //                audio is read and converted into this buffer which the caller
  //                can then use. When the FnAVFrame object is destroyed it will
  //                release the buffer.
  //
  // @return <0 if unsuccessful, otherwise the number of samples read.
  //
  int read(FnAVFrame& avFrame);

  ////////////////////////////////////////////////////////////////////////////////
  // close
  // Release all resources.
  //
  void close();

  /////////////////////////////////////////////////////////////////////////////////
  // setSampleFormat
  // Specify the desired output sample format of the reader.
  // This will configure an appropriate resampling context that is used in the
  // Read method to convert from a source format to the desired target format.
  //
  // @param avSampleFormat The desired target sample format.
  //
  // @return 0 if successful, non-zero otherwise.
  //
  int setSampleFormat(AVSampleFormat avSampleFormat);

  /////////////////////////////////////////////////////////////////////////////////
  // getSampleFormat
  // Return the target format of the reader.
  //
  // @return An AVSampleFormat enumeration of the target sample format.
  //
  AVSampleFormat getSampleFormat()
  {
    return _avSampleFormat;
  }

  ////////////////////////////////////////////////////////////////////////////////
  // getSampleRate
  // Return the sample rate of the audio in the source file.
  //
  // @return The sample rate if a file has been successfully opened and contains
  //         audio. Zero otherwise.
  //
  int getSampleRate();

  ////////////////////////////////////////////////////////////////////////////////
  // getNumberOfChannels
  // Return the channel count of the audio in the source file.
  //
  // @return The channel count if a file has been successfully opened and contains
  //         audio. Zero otherwise.
  //
  int getNumberOfChannels();

  ////////////////////////////////////////////////////////////////////////////////
  // getDuration
  // Return the duration of the audio stream in seconds.
  //
  // @return The duration of the audio stream in seconds.
  //
  double getDuration();

  ////////////////////////////////////////////////////////////////////////////////
  // setStartPosition
  // Set the start position of the audio reader. This is used to adjust the audio
  // with respect to video in the mov64Writer. This can be used to advance or
  // delay the audio. If the audio is advanced the start of the audio will not
  // be read. If the audio is delayed then silence is inserted.
  //
  // @param startTime The time in seconds of the position of the first read.
  //                  This value can be zero, positive or negative.
  //                  Time zero is the start of the file, all other values are
  //                  with respect to this position.
  //                  If positive, this will cause the reader to start reading
  //                  after of the start of the file. I.e. positive values will
  //                  skip the start of the file.
  //                  If negative, this will cause the reader to read 'ahead' of
  //                  the start of the file. As no data exists, it will provide
  //                  silence until the read position arrives at the start of the
  //                  file.
  //
  // @return 0 if successful,
  //         <0 otherwise.
  //
  int setStartPosition(double startTime);

private:
  AVFormatContext* _avFormatContext; // A format context is required for the file handling.
  AVStream* _avStreamAudio; // This provided all manipulation of the stream and codec context.
  AVFrame* _avFrameAudio; // This is provide so that it is initialized once and used many times. The decoded manages the read buffer.
  SwrContext* _swrContext;

  AVSampleFormat _avSampleFormat; // Desired target format.

  int64_t _readPosition;
};

////////////////////////////////////////////////////////////////////////////////
// mov64Writer
//
class mov64Writer : public DD::Image::Writer
{
private:
  enum WriterError { SUCCESS = 0, IGNORE_FINISH, CLEANUP };

public:
  explicit mov64Writer(DD::Image::Write* iop, bool componentOfMovWriter = false);
  ~mov64Writer();

  DD::Image::Knob* getCodecKnob() { return codecKnob_; }

  virtual bool movie() const { return true; }

  DD::Image::LUT* defaultLUT() const;
  bool open();
  void execute();
  void finish();
  bool isDefaultLUTKnob(DD::Image::Knob* knob) const;
  void knobs(DD::Image::Knob_Callback f);
  virtual int knob_changed(DD::Image::Knob* knob);

  static const Writer::Description d;

  const char* const* getCodecNames() const
  {
    return &codecsKnobLabels_.front();
  }

private:
  void freeFormat();
  AVColorTransferCharacteristic getColorTransferCharacteristic() const;
  AVPixelFormat                 getPixelFormat(AVCodec* videoCodec) const;
  AVOutputFormat*               initFormat(bool reportErrors) const;
  bool                          initCodec(AVOutputFormat* fmt, AVCodecID& outCodecId, AVCodec*& outCodec) const;

  void getPixelFormats(AVCodec*          videoCodec,
                       AVPixelFormat&    outNukeBufferPixelFormat,
                       AVPixelFormat&    outTargetPixelFormat,
                       int&              outBitDepth) const;

  int encodeVideo(AVCodecContext* avCodecContext, uint8_t* out, int outSize, const AVFrame* avFrame);
  void updateBitrateToleranceRange();
  bool isRec709Format(const int height) const;
  static bool IsYUV(AVPixelFormat pixelFormat);
  static bool IsYUVFromShortName(const char* shortName);
  static bool IsRGBFromShortName(const char* shortName);


  static int            GetPixelFormatBitDepth(const AVPixelFormat pixelFormat);
  static AVPixelFormat  GetPixelFormatFromBitDepth(const int bitDepth);
  static void           GetCodecSupportedParams(AVCodec* codec, bool& outLossyParams,
                                                bool& outInterGOPParams, bool& outInterBParams);

private:
  AVFormatContext*  formatContext_;
  AVStream* streamVideo_;
  AVStream* streamAudio_;
  AVStream* streamTimecode_;
  std::vector<std::string> formatsLongNames_;
  std::vector<const char*> formatsShortNames_;
  std::vector<const char*> codecsShortNames_;
  std::vector<const char*> codecsKnobLabels_;

  bool writeTimecode_; // Flag to add a timecode track to the movie.
  bool writeNCLC_; //Flag to add an NCLC atom to the mov header.
  bool isOpen_; // Flag for the configuration state of the FFmpeg components.
  WriterError error_;
  // knobs variables
  float fps_;
  int format_;
  int codec_;
  const char* _audioFile;
  float _audioOffset;
  int _audioOffsetUnit;
  int bitrate_;
  int bitrateTolerance_;
  int gopSize_;
  int bFrames_;
  int mbDecision_;
  int qMin_;
  int qMax_;
  bool componentOfMovWriter_;
  bool defaultLutUpdated_;
  DD::Image::Knob* codecKnob_;

#if defined(FN_LICENSED_PRORES_CODEC)
  // Apple ProRes codec specific.
  mutable const AppleProResCodecStr* appleProResCodec_; // Pointer into appleProResCodecLUT for the currently selected ProRes codec. Mutable is used as this is simply a pointer into an existing table of const data. The data is never modified. This pointer is used to reference the selected ProRes codec and unfortunately this happens inside a const function, so mutable is used to move the pointer to the appropriate table entry during configuration.
  const char* oldProResCodecName_; // This is a reference to the non SDK name, which is replaced with the correct codec profile name when creating a movie and restored when the movie is closed.
  std::auto_ptr<NukeCodecs::IEncoder> encoder_;
#endif // FN_LICENSED_PRORES_CODEC

  boost::shared_ptr<AudioReader> audioReader_;

  void configureAudioStream(AVCodec* avCodec, AVStream* avStream);
  void configureVideoStream(AVCodec* avCodec, AVStream* avStream);
  void configureTimecodeStream(AVCodec* avCodec, AVStream* avStream);
  AVStream* addStream(AVFormatContext* avFormatContext, enum AVCodecID avCodecId, AVCodec** pavCodec);
  int openCodec(AVFormatContext* avFormatContext, AVCodec* avCodec, AVStream* avStream);
  int writeAudio(AVFormatContext* avFormatContext, AVStream* avStream, bool flush);
  int writeVideo(AVFormatContext* avFormatContext, AVStream* avStream, bool flush);
  int writeTimecode(AVFormatContext* avFormatContext, AVStream* avStream, bool flush);
  int writeToFile(AVFormatContext* avFormatContext, bool finalise);
};


#endif // NUKE_MOV64WRITER_H

